/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.DeBlock
 *
 * de.bsvrz.kex.tls.osi7.DeBlock is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.DeBlock is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.DeBlock.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.DeBlock.
 *
 *  de.bsvrz.kex.tls.osi7.DeBlock ist Freie Software: Sie knnen es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder spteren
 *  verffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.DeBlock wird in der Hoffnung, dass es ntzlich sein wird, aber
 *  OHNE JEDE GEWHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewhrleistung der MARKTFHIGKEIT oder EIGNUNG FR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License fr weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.osi7;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.hexdump.HexDumper;

//~ KLASSEN ===================================================================

/**
 * TBD: Kurzbeschreibung TBD: Langebeschreibung
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kni (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DeBlock {

    /**
     * DebugLogger fr Debug-Ausgaben
     */
    private static final Debug debug = Debug.getLogger();

    //~ FELDER ================================================================

    /** Lnge des Anwendungsdatenbytes im DE-Block. Kann gem TLS zwischen 0 und 226 liegen. */
    private final int _countDeData;

    /** De-Kanal-Nummer des DE-Blcks. */
    private int _deChannel;

    /** Anwendungsdaten im DE-Block. */
    private TlsData _deData;

    /** Funktionsgruppe der DE-Blcke im EinzelTelegramm. */
    private final int _fg;

    /** Richtung/Anwendungs-ID im EinzelTelegramm. */
    private final int _id;

    /** OSI7 KnotenNummer (die ersten drei Byte eines Osi/Telegramms). */
    private final int _osi7Adress;

    /** Typ der DE-Daten. */
    private final int _typ;

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    /**
     * Erzeugt einen neues Objekt DeBlock aus den De-Block-Daten eines empfangen TLS-Telegramms. Dazu werden die TLS-Byte
     * eines empfangen De-Blocks (als TlsData Objekt gekapselt), die Fg und die Id des De-Block bergeben. Wird verwendet,
     * um aus einem empfangenen Telegramm die De-Blcke zu kapseln.
     *
     * @param deBlockData Die empfangen Tls-De-Block Daten als TlsData-Objekt
     * @param osi7Adress  Knotennummer der De des De-Blocks (aus dem Osi7Telegram)
     * @param fg          Funktionsgruppe dieses De-Blocks (aus dem Einzeltelegramm)
     * @param id          Id dieses De-Blocks (aus dem Einzeltelegramm)
     *
     * @throws IllegalOsi7TelegramStructureException
     *          Wenn formaler Aufbau des De-Blocks nicht korrekt (i.d.R. De-Block zu kurz oder zu lang).
     */
    public DeBlock(TlsData deBlockData, int osi7Adress, int fg, int id) throws IllegalOsi7TelegramStructureException {
        _osi7Adress  = osi7Adress;
        _fg          = fg;
        _id          = id;
        _deChannel   = deBlockData.readUnsignedByteAsInt();
        _typ         = deBlockData.readUnsignedByteAsInt();
        _countDeData = deBlockData.length() - 2;  // Anzahl der eigentlichen Anwendungsdaten ohne DE-Header

        if ((_countDeData < 0) || (_countDeData > 226)) {
            throw new IllegalOsi7TelegramStructureException(
                "DE-Block enthlt " + _countDeData +
                " Anwendungsbyte -> ausserhalb des gltigen Bereichs von 0..226 DE-Blcke!");
        }

        if (_countDeData > 0) {
            try {

                // Versuchen, die DE-Daten zu extrahieren...
                _deData = deBlockData.readNextTlsDataBlock(_countDeData);
            }
            catch (Exception e) {
                {  // DEBUG-BLOCK
                    String msg;

                    msg = "DE Anwendungsdaten nicht ermittelbar." + System.getProperty("line.separator");
                    msg += "Lnge des DE-Blcks grer als tatschliche Anzahl der Byte." +
                           System.getProperty("line.separator");
                    msg += HexDumper.toString(deBlockData.toBytes()) + System.getProperty("line.separator");
                    msg += "Exception: " + e;
                    debug.error(msg, e);
                }
            }
        } else {
            _deData = new TlsData(new byte[] {});
        }
    }

    /**
     * Erzeugt einen neues Objekt DeBlock aus den Anwendungsdaten eines zum Versand zu erzeugenden De-Blocks. Dazu werden
     * die Anwendungsdaten eines zu sendenden De-Blocks (also ohne die De-Block Headerinformationen) als TlsData Objekt
     * gekapselt, die Fg, Id, der Typ und die De-Kanalnummer bergeben. Wird verwendet, um einen De-Block zum versenden zu
     * erstellen.
     *
     * @param deApplicationData Die empfangen Tls-De-Block Daten als TlsData-Objekt
     * @param osi7Adress        Knotennummer der De des De-Blocks (aus dem Osi7Telegram)
     * @param fg                Funktionsgruppe dieses De-Blocks
     * @param id                Id dieses De-Blocks
     * @param typ               Typ dieses De-Blocks
     * @param deChannel         DeKanalnummer des De-Blocks
     *
     * @throws IllegalOsi7TelegramStructureException
     *          Wenn formaler Aufbau des De-Blocks nicht korrekt (i.d.R. De-Block zu kurz oder zu lang).
     */
    public DeBlock(TlsData deApplicationData, int osi7Adress, int fg, int typ, int id, int deChannel)
            throws IllegalOsi7TelegramStructureException {
        _osi7Adress  = osi7Adress;
        _fg          = fg;
        _id          = id;
        _typ         = typ;
        _deChannel   = deChannel;
        _countDeData = deApplicationData.length();  // Anzahl der eigentlichen Anwendungsdaten (ohne DE-Header)

        if ((_countDeData < 0) || (_countDeData > 226)) {
            throw new IllegalOsi7TelegramStructureException(
                "DE-Block enthlt " + _countDeData +
                " Anwendungsbyte -> ausserhalb des gltigen Bereichs von 0..226 DE-Blcke!");
        }

        if (_countDeData > 0) {
            try {

                // Versuchen, die DE-Daten zu extrahieren...
                _deData = deApplicationData.readNextTlsDataBlock(_countDeData);
            }
            catch (Exception e) {
                {  // DEBUG-BLOCK
                    String msg;

                    msg = "DE Anwendungsdaten nicht ermittelbar." + System.getProperty("line.separator");
                    msg += "Lnge des DE-Blcks grer als tatschliche Anzahl der Byte." +
                           System.getProperty("line.separator");
                    msg += HexDumper.toString(deApplicationData.toBytes()) + System.getProperty("line.separator");
                    msg += "Exception: " + e;
                    debug.error(msg, e);
                }
            }
        } else {
            _deData = new TlsData(new byte[] {});
        }
    }

    //~ GET METHODEN ==========================================================

    /**
     * Gibt die DE-Kanal-Nummer des DE-Blocks zurck.
     *
     * @return DE-Kanal-Nummer des DE-Blocks (0..255).
     */
    public int getDeChannel() {
        return _deChannel;
    }

    /**
     * Gibt die DE-Anwendungsdatenbytes des DE-Blocks zurck.
     *
     * @return DE-Anwendungsdatenbytes des DE-Blocks.
     */
    public TlsData getDeData() {
        return _deData;
    }

    /**
     * Liefert einen String auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe als String.
     *
     * @return String auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe ("Osi7Adresse-DeNummer-FG",
     *         z.B. "51043-193-4").
     */
    public String getDescriptionKey() {
        return ("" + this.getOsi7Adress() + "-" + this.getDeChannel() + "-" + this.getFg() + " [" +
                this.getOsi7AdressLocCode() + "-" + this.getDeChannel() + "-" + this.getFg() + "]");
    }

    /**
     * Gibt die Funktionsgruppe der DE-Blcke des Einzeltelgramms zurck.
     *
     * @return Funktionsgruppe der DE-Blcke des Einzeltelgramms (1..254).
     */
    public int getFg() {
        return _fg;
    }

    /**
     * Gibt die ID des DE-Blocks zurck.
     *
     * @return ID des DE-Blocks (1..255).
     */
    public int getId() {
        return _id;
    }

    /**
     * Liefert einen Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe als String.
     *
     * @return Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe ("Osi7Adresse-DeNummer-FG",
     *         z.B. "51043-193-4").
     */
    public String getKey() {
        return (String.valueOf(String.valueOf(this.getOsi7Adress() + "-" + this.getDeChannel() + "-" + this.getFg())));
    }

    /**
     * Liefert einen Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer, der Funktionsgruppe, des Typs und der Id als
     * String.
     *
     * @return Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer, der Funktionsgruppe, des Typs und der Id
     *         ("Osi7Adresse-DeNummer-FG-Typ-Id", z.B. "51043-193-4-29-2").
     */
    public String getLongKey() {
        return (String.valueOf(String.valueOf(this.getOsi7Adress() + "-" + this.getDeChannel() + "-" + this.getFg() +
                                              "-" + this.getTyp() + "-" + this.getId())));
    }

    /**
     * Liefert die KnotenNummer des Osi7Telegramms in Dezimaldarstellung (0..16777215).
     *
     * @return KnotenNummer des Osi7Telegramms in Dezimaldarstellung (0..16777215)
     */
    public int getOsi7Adress() {
        return _osi7Adress;
    }

    /**
     * Liefert die KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
     *
     * @return KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
     */
    public String getOsi7AdressLocCode() {
        return ((_osi7Adress >> 8) + "-" + (_osi7Adress & 0xff));
    }

    /**
     * Gibt einen De-Block gem TLS zurck.
     *
     * @return DE-Block als byte[] gem TLS.
     *         <pre>
     *                 Aufbau:
     *                 Byte[0] Lnge des De-Blocks
     *                 Byte[2] Daten-Endgerte
     *                 Byte[3] Typ der De-Daten
     *                 ----------------------------------------
     *                 Byte[4] 1. Anwendungsdatenbyte
     *                 .....
     *                 Byte[n] (n-3)-tes Anwendungsdatenbyte
     */
    public byte[] getTlsDeDataBlock() {
        byte[] tlsDeDataBlock = new byte[_deData.length() + 3];  // Lnge der Anwendungsdaten + Lngenbyte + De + Typ

        tlsDeDataBlock[0] = (byte) (tlsDeDataBlock.length - 1);  // Lngenbyte = Lnge ohne Lngenbyte
        tlsDeDataBlock[1] = (byte) getDeChannel();               // De-Kanal
        tlsDeDataBlock[2] = (byte) getTyp();                     // Typ der Daten

        byte[] data = _deData.toBytes();

        for (int i = 0; i < _deData.length(); ++i) {
            tlsDeDataBlock[i + 3] = data[i];  // Anwendungsdaten
        }

        return tlsDeDataBlock;
    }

    /**
     * Gibt den Typ des DE-Blocks zurck.
     *
     * @return Typ des DE-Blocks (1..255).
     */
    public int getTyp() {
        return _typ;
    }

    //~ SET METHODEN ==========================================================

    /**
     * Setzt die DE-Kanal-Nummer des DE-Blocks.
     *
     * @param deChannel Neue De-Kanalnummer fr diesen DeBlock.
     */
    public void setDeChannel(int deChannel) {
        _deChannel = deChannel;
    }

    //~ METHODEN ==============================================================

    /**
     * Ausgabe des De-Blocks mit Einzelinformationen und Anwendungsdaten Beispielausgabe:
     * <pre><code>
     * Knotennummer:1001 (256-1) DE:48 FG:6 Id:12 Typ:48 Lnge:3
     * 00000000:  02 7F FF                                              ..
     * </pre></code>
     *
     * @return String-Darstellung
     */
    public String toString() {
        StringBuffer sb      = new StringBuffer();
        final String NEWLINE = System.getProperty("line.separator");

        sb.append("Knotennummer:" + getOsi7Adress() + " (" + getOsi7AdressLocCode() + ")" + " DE:" + getDeChannel() +
                  " FG:" + getFg() + " Id:" + getId() + " Typ:" + getTyp() + " Lnge:" + getDeData().length());
        sb.append(NEWLINE);
        sb.append(HexDumper.toString(getDeData().toBytes()));

        return (sb.toString());
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 27.04.09
